// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
{**
   Implementations of Object, Tag, InvokeRecord and ORB.
 }
unit orb;

interface

{$I dorb.inc}

uses
  {$IFDEF WIN32}Windows,{$ENDIF}{$IFDEF LINUX}Libc,{$ENDIF}Classes,orb_int,orbtypes,
  SysUtils,disp_int,giop_int,addr_int,iior_int,iiop_int,req_int,value_int,map_int,
  env_int, code_int,utils,osthread,ir_int,comp_int;

type

  {** implements org.omg/CORBA/Object }
  TORBObject = class(TInterfacedObject,IORBObject)
  protected
    FIOR,FFWD_IOR : IIOR;
    FORB : IORB;
  protected
    function get_interface(): IInterfaceDef;
    function duplicate: IORBObject;
    procedure release;
    function _is_a(const repoid: string): Boolean;
    function non_existent: Boolean;
    function is_equivalent(obj: IORBObject): Boolean;
    function hash(max: _ulong): _ulong;
    procedure create_request(const ctx: IContext;
      const operation: Identifier; const arg_list: INVList;
      out request: IRequest; const flags: _ulong);
    function get_policy(const policy_type: PolicyType): IPolicy;
    function get_domain_managers: DomainManagersList;
    function set_policy_overrides(const policies: PolicyList;
      const set_add: SetOverrideType): IORBObject;

    // DORB extension
    procedure assign(const obj: IORBObject);
    function ior(): IIOR;
    function fwd_ior(): IIOR;
    function orb(): IORB;
    function narrow_helper(const repoid: string): Pointer; virtual;
    function is_forwarded(): Boolean;
    function _is_a_remote(const repoid: string): Boolean;
    procedure _unforward();
    procedure _forward(const obj :IORBObject);
    function repoid(): RepositoryID; virtual;
    function _request(const operation: string): IRequest; virtual;
    function refcount(): Longint;
  public
    constructor Create(); overload;
    constructor Create(ior : IIOR); overload;
    destructor Destroy; override;
  end;

{$IFNDEF USELIB}

  TORBState = (osNotRunning, osRunning, osServerShutdown, osClientShutdown, osDestroyed);

  {** implements org.omg/CORBA/ORB }
  TOrb = class(TORBObject,IORB)
  private
    FMessageCounter : MessageID;
    FAdapters : TStringList;
    FAdaptersLock: TRWLock;
    FShuttingdownAdapters: IInterfaceList;
    FInitRefs : TStringList;
{$IFDEF USE_ORB_CACHE}
    FCacheUsed: Boolean;
    FCacheRecord : IORBInvokeRecord;
{$ENDIF}
    FInvokes : ILongInterfaceMap;
    FInvokesLock: TRWLock;
    FDispatcher : ISelectDispatcher;
    FBindAddress : TStrings;
    FIORTempl : IIOR;
    //FRunning : Boolean;
    //FStopped : Boolean;
    FShutdown: boolean;
    FWaitForCompletion : Boolean;
    FValueFactories : TStringList;
    FValueFactoriesLock: TRWLock;
    FTheIDLock: TMutex;
    FInitRefsLock: TRecursiveMutex;
    FShutdownComplete: TThreadMonitor;
    FMainThread: TThreadHandle;
    FProperties: IProperties;
    FState: TORBState;
    function _get_invoke(id: ORBMessageID):IORBInvokeRecord;
    procedure validateState();
    procedure blockServerShutdownComplete();
  protected
    function object_to_string(const obj: IORBObject): string;
    function string_to_object(const str: string): IORBObject;
    // Dynamic Invocation related operations
    procedure create_list(const count: longint; out new_list: INVList);
    procedure create_operation_list(const oper: IOperationDef; out new_list: INVList);
    procedure get_default_context(out ctx: IContext);
    procedure send_multiple_requests_oneway(const req: RequestSeq);
    procedure send_multiple_requests_deferred(const req: RequestSeq);
    function poll_next_response: Boolean;
    procedure get_next_response(out req: IRequest);
    // Service information operations
    function get_service_information(const service_type: ServiceType;
      out service_information: ServiceInformation): Boolean;
    function list_initial_references: ObjectIDList;
    // Initial reference information
    function resolve_initial_reference(const id: ObjectID): IORBObject;
    // TypeCode creation operations
    function create_struct_tc(const id: RepositoryID; const name : Identifier;
      const members: StructMemberSeq): ITypeCode;
    function create_union_tc(const id: RepositoryID; const name : Identifier;
      const discriminator_type: ITypeCode; const members: UnionMemberSeq): ITypeCode;
    function create_enum_tc(const id: RepositoryID; const name : Identifier;
      const members: EnumMemberSeq): ITypeCode;
    function create_alias_tc(const id: RepositoryID; const name : Identifier;
      const original_type: ITypeCode): ITypeCode;
    function create_exception_tc(const id: RepositoryID; const name : Identifier;
      const members: StructMemberSeq): ITypeCode;
    function create_interface_tc(const id: RepositoryID; const name : Identifier): ITypeCode;
    function create_string_tc(const bound: _ulong): ITypeCode;
    function create_wstring_tc(const bound: _ulong): ITypeCode;
    function create_fixed_tc(const digits: _ushort; const scale: short): ITypeCode;
    function create_sequence_tc(const bound: _ulong; const element_type : ITypeCode): ITypeCode;
    function create_array_tc(const length: _ulong; const element_type : ITypeCode): ITypeCode;
    function create_value_tc(const id: RepositoryID; const name : Identifier;
      const type_modifier: TValueModifier; const concrete_base: ITypeCode;
      const members: ValueMemberSeq): ITypeCode;
    function create_value_box_tc(const id: RepositoryID; const name : Identifier;
      const boxed_type: ITypeCode): ITypeCode;
    function create_native_tc(const id: RepositoryID; const name : Identifier): ITypeCode;
    function create_recursive_tc(const id: RepositoryID): ITypeCode;
    function create_abstract_interface_tc(const id: RepositoryID; const name : Identifier): ITypeCode;
    // Thread related operations
    function work_pending: Boolean;
    procedure perform_work;
    procedure run;
    procedure shutdown(wait_for_completion: Boolean);
    procedure _destroy(); // CORBA::ORB:destroy()
    // Policy related operations
    function create_policy(_type: PolicyType; const val: IAny): IPolicy;
    // Value factory operations
    function register_value_factory(const id: RepositoryID;
      const factory: IValueFactory): IValueFactory;
    procedure unregister_value_factory(const id : RepositoryID);
    function lookup_value_factory(const id : RepositoryID): IValueFactory;
    // DORB extension
    procedure answer_bind(id: ORBMessageID;stat: TLocateStatus;obj : IORBObject);
    procedure answer_invoke(id: ORBMessageID; stat: TInvokeStatus; obj: IORBObject; req: IORBRequest);
    procedure answer_shutdown(const oa: IObjectAdaptor);
    procedure bindaddrs(List : TStrings);
    function bind(const repoid: string; const addr: string = ''): IORBObject; overload;
    function bind(const repoid: string; const oid: IObjectTag; const addr: string = ''): IORBObject; overload;
    function bind_addr(const repoid: string; const oid: IObjectTag; const addr: IAddress; out obj: IORBObject): TLocateStatus;
    function bind_async(const repoid: string; const oid: IObjectTag; const addr: IAddress; const cb: IORBCallback; id : ORBMessageID): ORBMessageID;
    function BOA_Init(): IUnknown;
    function create_environment: IEnvironment;
    function dispatcher(): ISelectDispatcher;
    function get_bind_reply(id: ORBMessageID; out obj: IORBObject): TLocateStatus;
    function get_invoke_reply (id: ORBMessageID;out obj: IORBObject; out req: IORBRequest): TInvokeStatus;
    function ior_to_object(const Ior: IIOR): IORBObject;
    function ior_template: IIOR;
    function invoke_async(const obj: IORBObject; const req: IORBRequest; const pr: IPrincipal; resp: Boolean = true; cb: IORBCallBack = nil; orbmesid: ORBMessageID = nil): ORBMessageID;
    function is_impl(const obj: IORBObject): Boolean;
    function new_MessageID(): MessageID;
    function _non_existent(obj: IORBObject): Boolean;
    procedure register_oa(const oa: IObjectAdaptor; prio: integer);
    function request_type(id: ORBMessageID): TRequestType;
    procedure set_initial_reference(const id: string; const obj: IORBObject);
    function string_to_tag(const str: string): IObjectTag;
    procedure unregister_oa(const oa: IObjectAdaptor);
    function wait(id: ORBMessageID; timeout : long = -1): Boolean;
    procedure cancel(id: ORBMessageID); overload;
    procedure cancel(mid: MessageID); overload;
    function create_named_value: TNamedValue;
    function get_current_invoke_rec(): IORBInvokeRecord;
    function get_current_msgid(): MessageID;
    procedure set_Properties(const props: IProperties);
    function get_Properties(): IProperties;
  protected
    procedure do_shutdown();
    function get_invoke(mid: MessageID):IORBInvokeRecord;
    procedure del_invoke(id: MessageID);
    function builtin_invoke(mid: ORBMessageID; const obj: IORBObject; const req: IORBRequest):Boolean;
    function is_local(const obj: IORBObject): Boolean;
    function get_oa(const obj: IORBObject): IObjectAdaptor;
    function create_invoke(msgid: MessageID): IORBInvokeRecord;
    procedure add_invoke(const inv: IORBInvokeRecord);
    function http_to_object(const str: string): IORBObject;
    function file_to_object(const str: string): IORBObject;
    function _get_interface(const obj: IORBObject): IInterfaceDef;
    function get_msgid(rec: ORBMessageId): MessageID;
    function get_orbid(msgid: MessageId): ORBMessageID;
    function get_InvokeHint(id: ORBMessageId): Pointer;
    procedure set_InvokeHint(id: ORBMessageId; hint: Pointer);
    procedure set_RequestHint(id: ORBMessageId; hint: Pointer);
    function get_RequestHint(id: ORBMessageId): Pointer;
    function new_orbid(msgid: MessageId = 0): ORBMessageId;
  public
    constructor Create();
    destructor Destroy(); override;
  end;

  TORBInvokeRecord = class(TORBObject,IORBInvokeRecord)
  private
    FORB: IORB;
    FMessageID : MessageID;
    FObject: IORBObject;
    FRequest : IORBRequest;
    FResponse_req : Boolean;
    FCallBack : IORBCallBack;
    FAdapter : IObjectAdaptor;
    FType : TRequestType;
    FStat: TInvokeStatus;
    FLocateStatus: TLocateStatus;
    FHaveResult : Boolean;
    FActive: Boolean;
    FTag : IObjectTag;
    FAddress : IAddress;
    FReqHint,
    FInvHint: Pointer;
    FPrincipal: IPrincipal;
  protected
    { IORBInvokeRecord }
    procedure init_invoke(const orb: IORB; const obj: IORBObject; const pr: IPrincipal; const req: IORBRequest; resp: Boolean; const cb: IORBCallBack);
    procedure init_bind(const orb: IORB; const repoid: string; const tag: IObjectTag; const addr: IAddress; const cb: IORBCallBack; const oa: IObjectAdaptor);
    procedure set_oa(const oa: IObjectAdaptor);
    function id: MessageID;
    function completed(): Boolean;
    function get_answer_invoke(out state: TInvokeStatus;out obj: IORBObject;out req: IORBRequest): Boolean;
    procedure set_answer_invoke(const state: TInvokeStatus;const obj: IORBObject;const req: IORBRequest);
    procedure set_answer_bind(const state: TLocateStatus;const obj: IORBObject);
    procedure oa(oa: IObjectAdaptor);
    function request_type: TRequestType;
    procedure clean();
    function callback(): IORBCallBack;
    function get_answer_bind(out state: TLocateStatus;out obj: IORBObject): Boolean;
    function get_oa(): IObjectAdaptor;
    function active(): boolean;
    procedure deactivate();
    procedure set_RequestHint(hint: Pointer);
    procedure set_InvokeHint(hint: Pointer);
    function get_RequestHint(): Pointer;
    function get_InvokeHint(): Pointer;
    function get_principal(): IPrincipal;
  public
    constructor Create(msgId: MessageID);
  end;

{$ENDIF}

  TObjectTag = class(TInterfacedObject,IObjectTag)
  private
    FName: string;
  protected
    function get_length: _ulong;
    function get_bytes: string;
    function equals(const val: string): Boolean;
  public
    constructor Create(Name: string);
  end;

  TORBAsyncCallback = class(TInterfacedObject, IORBCallBack, IORBasyncCallback)
  private
    FNotified: boolean;
    FCond: TCondVar;
    FCondMutex: TRecursiveMutex;
  protected
    function waitfor(const orb: IORB; mid: ORBMessageID; event: TORBCallbackEvent; tmout: Longint = -1): boolean;
    procedure notify(const orb: IORB; mid: ORBMessageID; event: TORBCallbackEvent);
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TPrincipalCurrent = class
  public
    class function _narrow(const obj : IORBObject): IPrincipalCurrent;
  end;

  TPrincipalCurrent_impl = class(TORBObject, IPrincipalCurrent)
  protected
    function narrow_helper(const repoid: string): Pointer;  override;
    function get_principal(): IPrincipal;
  end;

  TProperties = class(TInterfacedObject, IProperties)
  private
    FProps: TStrings{List};
    procedure LoadDefaultProperties;
  protected
    { IProperties }
    procedure set_Property(const key, value: string);
    function get_Property(const key: string): string;
  public
    constructor Create(props: TStrings);
    destructor Destroy; override;
  end;

{$IFDEF USELIB}
  function ORB_Init(props: TStrings): IOrb; external MTDORB_Library_Name;
  function ORB_Instance: IORB;  external MTDORB_Library_Name;
{$ELSE}
  function ORB_Init(props: TStrings): IOrb;
  function ORB_Instance: IORB;
{$ENDIF}

implementation

uses
  static,ior,iop,address,disp,proxy,iiop,boa,imr,
  mcomp_int,orbcomp,codeset,exceptions,except_int,map,
  poa_int, poa,throw,imr_impl,poa_impl,env,intercept,mtmanager,
  tcode,policies,policy_int,req,ir,any,ssl,osnet;

{$IFNDEF USELIB}
var
  ORBInstance : IORB;
threadvar
  currentMsgId: MessageID;
{$ENDIF}

{$IFNDEF USELIB}
function ORB_Instance: IORB;
begin
  result := ORBInstance;
end;

{**
  ORB initialization routine.
  Initializes sockets and creates instance of ORB.
}
function ORB_Init(props: TStrings): IOrb;
var
  i : integer;
  bindaddr,iiopaddr : TStrings;
  pvalue,naming_addr,naming_ior : string;
  tag : IObjectTag;
  naming : IORBObject;
  iiopserv: Boolean;
  server: IIIOPServer;
  mcprof : IMultiComponentProfile;
  mc : IMultiComponent;
  addr: IAddress;
  retval: boolean;
  properties: IProperties;
  giopver: _ushort;
  giopver_str: string;
begin
  TOSNet.sockinit();
  iiopserv := true;
  naming_addr := '';
  naming_ior := '';
  giopver_str := '1.0';
  if ORBInstance = nil then
    begin
      Codeset.CodesetInit;
{$IFDEF HAVE_SSL}
      Ssl.SslInit(props);
{$ENDIF}
      bindaddr := TStringList.Create;
      iiopaddr := TStringList.Create;
      try
        properties := TProperties.Create(props);
        ORBInstance := TORB.Create;
        ORBInstance.set_Properties(properties);
{$IFDEF HAVE_THREADS}
        thread_setup();
{$ENDIF}
        //interceptors
        retval := TInitInterceptor._exec_initialize(ORBInstance, 'dorb-local-orb', props);
        if not retval then
          dorb_throw(st_INITIALIZE);
        i := 0;
        if (props <> nil) then
          while i < props.count do
            begin
              pvalue := Uppercase(props[i]);
              if pvalue = '-ORBBINDADDR' then
                begin
                  inc(i);
                  bindaddr.Add(props[i]);
                end
              else if pvalue = '-ORBIIOPADDR' then
                begin
                  inc(i);
                  iiopaddr.Add(props[i]);
                end
              else if pvalue = '-ORBNOIIOPSERVER' then
                iiopserv := false
              else if pvalue = '-ORBNAMINGIOR' then
                begin
                  inc(i);
                  naming_ior := props[i];
                end
              else if pvalue = '-ORBNAMINGADDR' then
                begin
                  inc(i);
                  naming_addr := props[i];
                end
              else if pvalue = '-ORBGIOPVERSION' then
                begin
                  inc(i);
                  giopver_str := props[i];
                end;
              inc(i);
            end;

        if (Length(giopver_str) <> 3) or
           (giopver_str[1] <> '1') or
           (giopver_str[2] <> '.') or
           (giopver_str[3] < '0') or
           (giopver_str[3] > '2') then
          dorb_throw(st_INITIALIZE);
        giopver := ((Ord(giopver_str[1]) - Ord('0')) shl 8) or (Ord(giopver_str[3]) - Ord('0'));
        // create IIOP client
        ORBInstance.dispatcher.block(false);
        TIIOPProxy.Create(ORBInstance, giopver);

        ORBInstance.bindaddrs(bindaddr);
        if iiopserv then
          begin
            server := TIIOPServer.Create(ORB_Instance);
            if iiopaddr.Count = 0 then
              server.listen()
            else
              for i := 0 to Pred(iiopaddr.count) do
                begin
                  addr := TInetAddress.parse(iiopaddr[i]);
                  if addr = nil then
                    dorb_throw(st_BAD_PARAM, MinorBadAddress, COMPLETED_NO);
                  if not server.listen_addr(addr,nil) then
                    dorb_throw(st_INITIALIZE);
                end;
          end;
        // connect to naming service
        if naming_addr <> '' then
          begin
            tag := ORB_Instance.string_to_tag('NameService');
            naming := orb_instance.bind('IDL:omg.org/CosNaming/NamingContext:1.0',tag,naming_addr);
            if naming <> nil then
              orb_instance.set_initial_reference('NameService',naming);
          end
        else if naming_ior <> '' then
          begin
            tag := ORB_Instance.string_to_tag('NameService');
            naming := orb_instance.string_to_object(naming_ior);
            if naming <> nil then
              orb_instance.set_initial_reference('NameService',naming);
          end;
        if not TCodeset.disabled then begin
          mc := TMultiComponent.Create() as IMultiComponent;
          mc.set_id(TAG_CODE_SETS);
          //mc.add_component(TCodesetComponent.Create($00010001,$00010109,nil,nil) as IORBComponent);
          mc.add_component(TCodesetComponent.Create(TCodeset.special_cs(NativeCS).id,TCodeset.special_cs(NativeWCS).id,nil,nil) as IORBComponent);
          mcprof := TMultiCompProfile.Create(mc,TAG_MULTIPLE_COMPONENTS);
          ORB_Instance.ior_template.add_profile(mcprof);
        end;
      finally
        bindaddr.free;
        iiopaddr.free;
      end;
    end;
  result := ORBInstance;
end;
{$ENDIF}

//***************************************************************
//  TORBObject
//***************************************************************
constructor TORBObject.Create();
begin
  FORB := ORB_Instance;
end;

constructor TORBObject.Create(ior : IIOR);
begin
  FIOR := ior;
  FORB := ORB_Instance;
  FFWD_IOR := nil;
end;

destructor TORBObject.Destroy;
begin
  FIOR := nil;
  FORB := nil;
  FFWD_IOR := nil;
  inherited;
end;

{**  Returns Interoperable Object Reference }
function TORBObject.IOR(): IIOR;
begin
  result := FIOR;
end;

function TORBObject.refcount: Longint;
begin
  result := FRefCount;
end;

function TORBObject.fwd_ior(): IIOR;
begin
  if assigned(FFWD_IOR) then
    result := FFWD_IOR
  else
    result := FIOR;
end;

function TORBObject._request(const operation: string): IRequest;
begin
  result := CreateRequest(self,operation);
end;

function TORBObject.repoid(): RepositoryID;
begin
  result := FIOR.objid;
end;

function TORBObject._is_a(const repoid: string): Boolean;
begin
  result := true;
  if strcomp(PChar(repoid),'IDL:omg.org/CORBA/Object:1.0')=0 then exit;
  if strcomp(PChar(repoid),PChar(self.repoid))=0 then exit;
  if narrow_helper(repoid)<>nil then exit;
  result := false;
end;

function TORBObject.non_existent: Boolean;
begin
  result := FORB._non_existent(self);
end;

function TORBObject._is_a_remote(const repoid: string): Boolean;
begin
  if FORB.is_impl(self) then
    result := false
  else
    result := true; // must be rewritten 
end;

{**  Returns current ORB instance }
function TORBObject.ORB(): IORB;
begin
  result := FORB;
end;

function TORBObject.is_forwarded(): Boolean;
begin
  result := FFWD_IOR <> nil;
end;

procedure TORBObject._unforward();
begin
  FFWD_IOR := nil;
end;

procedure TORBObject._forward(const obj :IORBObject);
begin
  FFWD_IOR := CreateIOR_ior(obj.IOR);
end;

{**  Gets common properties from specified object}
procedure TORBObject.assign(const obj: IORBObject);
begin
  FIOR := CreateIOR_ior(obj.IOR);
  FORB := obj.ORB;
end;

function TORBObject.narrow_helper(const repoid: string): Pointer;
begin
  result := nil;
end;

procedure TORBObject.create_request(const ctx: IContext;
  const operation: Identifier; const arg_list: INVList;
  out request: IRequest; const flags: _ulong);
begin
  assert(false,'not imlplemented');
end;

function TORBObject.duplicate: IORBObject;
begin
  assert(false,'not imlplemented');
end;

function TORBObject.get_domain_managers: DomainManagersList;
begin
  assert(false,'not imlplemented');
end;

function TORBObject.get_interface: IInterfaceDef;
begin
  result := FORB._get_interface(self)
end;

function TORBObject.get_policy(const policy_type: PolicyType): IPolicy;
begin
  // not implemented
  result := nil;
  raise SystemException.Create_type(st_BAD_PARAM);
end;

function TORBObject.hash(max: _ulong): _ulong;
begin
  result := utils.string_hash(FIOR.stringify,max)
end;

function TORBObject.is_equivalent(obj: IORBObject): Boolean;
begin
  result := true;
  if obj = (self as IORBObject) then exit;
  result := false; // must be rewritten
end;

procedure TORBObject.release;
begin
  assert(false,'not imlplemented');
end;

function TORBObject.set_policy_overrides(const policies: PolicyList;
  const set_add: SetOverrideType): IORBObject;
begin
  result := nil;
  assert(false,'not imlplemented');
end;

{$IFNDEF USELIB}
//*************************************************************
//  TOrb
//**************************************************************
constructor TOrb.Create();
begin
  FORB := nil;
  FAdapters := TStringList.Create();
  FInitRefs := TStringList.Create();
{$IFDEF USE_ORB_CACHE}
  FCacheUsed := false;
  FCacheRecord := TORBInvokeRecord.Create();
{$ENDIF}
  FInvokes := TLongInterfaceMap.Create;
  FDispatcher := TSelectDispatcher.CreateEx;
  FBindAddress := TStringList.Create;
  FValueFactories := TStringList.Create;
  FIORTempl := TIOR.Create;
  //FRunning := False;
  //FStopped := False;
  FShutdown := false;
  FWaitForCompletion := false;
  FAdaptersLock := TRWLock.Create();
  FInvokesLock := TRWLock.Create();
  FValueFactoriesLock := TRWLock.Create();
  FTheIDLock := TMutex.Create();
  FInitRefsLock := TRecursiveMutex.Create();
  FShutdownComplete := TThreadMonitor.Create;
  FState := osNotRunning;
end;

destructor TOrb.Destroy();
begin
  FAdapters.Free;
  FValueFactories.Free;
  FInitRefs.Free;
  FInvokesLock.wrlock();
  try
    FInvokes := nil;
  finally
    FInvokesLock.wrunlock();
  end; { try/finally }
  FBindAddress.Free;
  FInitRefsLock.Free;
  FTheIDLock.Free;
  FValueFactoriesLock.Free;
  FInvokesLock.Free;
  FAdaptersLock.Free;
  FShutdownComplete.Free;
  inherited Destroy();
end;

procedure TOrb.cancel(id: ORBMessageID);
var
  rec : IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  if rec <> nil then
    begin
      rec.deactivate();
      if rec.get_oa <> nil then
        rec.get_oa.cancel(rec);
{$IFNDEF HAVE_THREADS}
      del_invoke(rec.id());
{$ENDIF}      
    end;
end;

function TOrb.get_InvokeHint(id: ORBMessageId): Pointer;
var
  rec: IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  Assert(rec <> nil);
  //rec._AddRef;
  result := rec.get_InvokeHint;
end;

function TOrb.get_msgid(rec: ORBMessageId): MessageID;
begin
  result := rec.id();
end;

function TOrb.get_orbid(msgid: MessageId): ORBMessageID;
begin
  result := get_invoke(msgid);
end;

function TOrb.get_RequestHint(id: ORBMessageId): Pointer;
var
  rec: IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  Assert(rec <> nil);
  result := rec.get_RequestHint();
end;

function TOrb.new_orbid(msgid: MessageId): ORBMessageId;
begin
  if msgid = 0 then msgid := new_MessageID;
  result := create_invoke(msgid);
end;

procedure TOrb.set_InvokeHint(id: ORBMessageId; hint: Pointer);
var
  rec: IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  Assert(rec <> nil);
  rec.set_InvokeHint(hint);
end;

procedure TOrb.set_RequestHint(id: ORBMessageId; hint: Pointer);
var
  rec: IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  Assert(rec <> nil);
  rec.set_RequestHint(hint);
end;

procedure TORB.do_shutdown();
var
  i : integer;
  OAList : IInterfaceList;
  oa: IObjectAdaptor;
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FShutdownComplete);
  // If the FShutdown is false, or the server side has already
  // shutdown then do nothing
  if not FShutdown or (FState = osServerShutdown) then
    Exit;
  Assert((FState <> osClientShutdown) and (FState <> osDestroyed));
  // If run was called then only the main thread may complete the
  // shutdown
  Assert((FState = osNotRunning) or (FMainThread = TORBThread.CurrentThread()));
  FShuttingDownAdapters := TInterfaceList.Create;
  OAList := TInterfaceList.Create;
  FAdaptersLock.rdlock();
  try
    //for i := 0 to Pred(FAdapters.count) do
    for i := Pred(FAdapters.count) downto 0 do
      begin
        oa := IObjectAdaptor(Pointer(FAdapters.Objects[i]));
        //FShuttingDownAdapters.Add(IObjectAdaptor(Pointer(FAdapters.Objects[i])));
        //OAList.Add(IObjectAdaptor(Pointer(FAdapters.Objects[i])));
        FShuttingDownAdapters.Add(oa);
        OAList.Add(oa);
        FAdapters.Delete(i);
        oa._Release;
        oa := nil;
      end;
  finally
    FAdaptersLock.rdunlock();
  end; { try/finally }
  for i := 0 to Pred(OAList.count) do begin
    with IObjectAdaptor(OAList[i]) do
      shutdown(FWaitForCompletion);
  end;
  if FWaitForCompletion then
    while FShuttingDownAdapters.Count > 0 do
      FDispatcher.run(false);
  OAList := nil;
  FShuttingDownAdapters := nil;
  // terminate ORB threads
  thread_shutdown();

  // Mark the server side state as shutdown and notify any waiting
  // threads
  FState := osServerShutdown;

  FShutdownComplete.notifyAll();
end;

procedure TORB.cancel(mid: MessageID);
var
  rec : IORBInvokeRecord;
begin
  rec := get_invoke(mid);
  if rec <> nil then
    begin
      rec.deactivate();
      if rec.get_oa <> nil then
        rec.get_oa.cancel(rec);
{$IFDEF HAVE_THREADS}
      del_invoke(rec.id());
{$ENDIF}
      currentMsgId := 0;      
    end;
end;

function TORB._non_existent(obj: IORBObject): Boolean;
var
  req : IRequest;
begin
  result := true;
  req := obj._request('_non_existent');
  req.get_result.argument.set_type(_tc_boolean);
  try
    req.invoke;
  except
  end;
  if req.get_environment.get_exception <> nil then exit;
  assert(req.get_result.argument.to_boolean(result));
end;

procedure TORB.answer_shutdown(const oa: IObjectAdaptor);
var
  i : integer;
begin
  for i := 0 to Pred(FShuttingDownAdapters.count) do
    if IObjectAdaptor(FShuttingDownAdapters[i]) = oa then
      begin
        FShuttingDownAdapters.Delete(i);
        break;
      end;
end;

procedure TORB.shutdown(wait_for_completion: Boolean);
var
  obj: IORBObject;
  pc: IPOACurrent;
  inInvocation: boolean;
begin
  // If wait_for_completion is true then find out whether we're inside
  // a method invocation -- if so throw a BAD_INV_ORDER exception
  if wait_for_completion then begin
    inInvocation := false;
    try
      obj := resolve_initial_reference('POACurrent');
      pc := TPOACurrent._narrow(obj);
      inInvocation := (pc <> nil) and {pc.inupcall()}pc.iscurrent();
    except
    end; { try/except }
    pc := nil;
    obj := nil;
    if inInvocation then
      dorb_throw(st_BAD_INV_ORDER);
  end;
  FWaitForCompletion := wait_for_completion;
  FShutdown := true;
  FDispatcher.interrupt_dispatch();

  // waitForCompletion false? We're done.
  if not FWaitForCompletion then
    Exit;
  // If run was called and this is not the main thread and
  // waitForCompletion is true then wait for the shutdown to
  // complete.
  if (FState = osRunning) and (FMainThread <> TORBThread.CurrentThread()) then begin
    blockServerShutdownComplete();
    Exit;
  end;

  // This is the main thread -- complete the shutdown process
  do_shutdown();
end;

function TORB.builtin_invoke(mid: ORBMessageID; const obj: IORBObject; const req: IORBRequest):Boolean;
begin
  //  not imlplemented
  result := false;
end;

function TORB.is_local(const obj: IORBObject): Boolean;
var
  p1,p2 : IIORProfile;
begin
  result := true;
  if obj.ior = nil then
    exit;
  // GIOPSimpleProf
  if obj.ior.profile(0).id = 10000 then
    exit;
  p1 := obj.ior.profile(0);
  p2 := FIORTempl.profile(0);
  if p1.address.compare(p2.address) then
    exit;
  result := false;
end;

function TORB.is_impl(const obj: IORBObject): Boolean;
var
  oa : IObjectAdaptor;
begin
  result := false;
  if obj = nil then exit;
  oa := get_oa(obj);
  result := (oa <> nil) and oa.is_local and (oa.skeleton(obj) = obj);
end;

function TORB.BOA_Init(): IUnknown;
var
  i : integer;
  obj : IORBObject;
begin
  FAdaptersLock.rdlock();
  try
    for i := 0 to Pred(FAdapters.Count) do
      begin
        if strcomp('dorb-local-boa',PChar(IObjectAdaptor(Pointer(FAdapters.Objects[i])).get_oaid))=0 then
          begin
            result := IObjectAdaptor(Pointer(FAdapters.Objects[i]));
            exit;
          end;
      end;
  finally
    FAdaptersLock.rdunlock();
  end; { try/finally }
  result := TBOAImpl.Create(self);
  obj := resolve_initial_reference('LocalImplementationRepository'); // check imr
end;

function TORB.ior_template: IIOR;
begin
  result := FIORTempl;
end;

procedure TORB.set_initial_reference(const id: string; const obj: IORBObject);
var
  i : integer;
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FInitRefsLock);
  i := FInitRefs.IndexOf(id);
  if i >= 0 then
    begin
      IORBObject(Pointer(FInitRefs.Objects[i]))._Release;
      FInitRefs.Delete(i);
    end;
  if obj <> nil then
    begin
      obj._AddRef;
      FInitRefs.AddObject(id,Pointer(obj));
    end;
end;

{**
  Adds bind addresses into address list
}
procedure TORB.bindaddrs(List : TStrings);
begin
  FBindAddress.AddStrings(List);
end;

function TORB.string_to_tag(const str: string): IObjectTag;
begin
  result := TObjectTag.Create(str);
end;

function TORB.get_bind_reply(id: ORBMessageID; out obj: IORBObject): TLocateStatus;
var
  rec : IORBInvokeRecord;
  state : TLocateStatus;
begin
  result := orbtypes.LocateUnknown;
  rec := _get_invoke(id);
  if rec = nil then exit;
  rec.get_answer_bind(state,obj);
  del_invoke(rec.id());
  result := state;
end;


function TORB.bind_async(const repoid: string; const oid: IObjectTag; const addr: IAddress;
  const cb: IORBCallback; id : ORBMessageID): ORBMessageID;
var
  rec: IORBInvokeRecord;
  i : integer;
  cback: IORBCallback;
begin
  rec := _get_invoke(id);
  if rec = nil then
    rec := new_orbid();
  cback := cb;
  if cback = nil then
    cback := TORBAsyncCallback.Create();
  rec.init_bind(self,repoid,oid,addr,cback,nil);
  add_invoke(rec);
  FAdaptersLock.rdlock();
  try
    for i := 0 to Pred(FAdapters.Count) do
      begin
        rec.oa(IObjectAdaptor(Pointer(FAdapters.Objects[i])));
        if IObjectAdaptor(Pointer(FAdapters.Objects[i])).bind(rec,repoid,oid,addr) then
          begin
            result := rec;
            exit;
          end;
      end;
  finally
    FAdaptersLock.rdunlock();
  end; { try/finally }
  answer_bind(rec,LocateUnknown,nil);
  result := rec;
end;


function TORB.bind_addr(const repoid: string; const oid: IObjectTag;
  const addr: IAddress; out obj: IORBObject): TLocateStatus;
var
  id: ORBMessageID;
begin
  result := orbtypes.LocateHere;
  id := bind_async(repoid,oid,addr,nil,nil);
  if id = nil then
    begin
      result := orbtypes.LocateUnknown;
      exit;
    end;
  if wait(id) then
    result := get_bind_reply(id,obj);
end;

function TORB.bind(const repoid: string; const addr: string): IORBObject;
begin
  result := bind(repoid,TObjectTag.Create('') as IObjectTag,addr);
end;

function TORB.bind(const repoid: string; const oid: IObjectTag; const addr: string): IORBObject;
var
  addrs: TStrings;
  address: IAddress;
  i : integer;
  stat : TLocateStatus;
  obj : IORBObject;
begin
  result := nil;
  addrs := TStringList.Create;
  try
    if addr = '' then
      addrs.AddStrings(FBindAddress)
    else
      addrs.Add(addr);
    for i := 0 to pred(addrs.count) do
      begin
        address := TInetAddress.parse(addrs[i]);
        if address <> nil then
          begin
            stat := bind_addr(repoid,oid,address,obj);
            if stat = orbtypes.LocateHere then
              begin
                result := obj;
                exit;
              end;
          end;
      end;
  finally
    addrs.Free;
  end;
end;

procedure TORB.answer_bind(id: ORBMessageID;stat: TLocateStatus;obj : IORBObject);
var
  rec : IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  if rec <> nil then
    begin
      rec.set_answer_bind(stat,obj);
      //if assigned(rec.callback()) then
        //rec.callback.callback(self,rec.id,_Bind);
    end;
end;

{**
  Sets answer to record with specified message id
**}
procedure TORB.answer_invoke(id: ORBMessageID; stat: TInvokeStatus; obj: IORBObject; req: IORBRequest);
var
  rec : IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  if rec <> nil then
    begin
      rec.set_answer_invoke(stat,obj,req);
      //if rec.callback <> nil then
        //rec.callback.callback(self,mid,_Invoke);
    end;
end;

{**
  Returns request type of invokation record with
  specified message id;
}
function TORB.request_type(id: ORBMessageID): TRequestType;
var
  rec : IORBInvokeRecord;
begin
  rec := _get_invoke(id);
  if rec <> nil then
    result := rec.request_type
  else
    result := orbtypes.RequestUnknown;
end;

{$IFDEF WIN32}
function ConsoleCtrlHandler(ctrlType: DWORD): BOOL; stdcall;
begin
  case ctrlType of
    CTRL_BREAK_EVENT,
    CTRL_C_EVENT,
    CTRL_CLOSE_EVENT: begin
      if ORBInstance <> nil then
        ORBInstance.shutdown(false{true});
      result := true;
    end;
    else result := false;
  end;
end;
{$ENDIF}
{$IFDEF LINUX}
  //Signal(SIGHUP, SIGINT, SIGTERM, SIGQUIT);
procedure ConsoleCtrlHandler(sig: integer); cdecl;
begin
  if ORBInstance <> nil then
    ORBInstance.shutdown(false);
end;
{$ENDIF}

procedure TORB.run();
begin
  validateState;
  if FMainThread <> TORBThread.CurrentThread() then begin
    // If this is not the main thread then block until the ORB is
    // shutdown
    blockServerShutdownComplete();
    Exit;
  end;
{$IFDEF WIN32}
  SetConsoleCtrlHandler(@ConsoleCtrlHandler, true);
{$ENDIF}
{$IFDEF LINUX}
  //signal(SIGHUP, ConsoleCtrlHandler);
  //signal(SIGINT, ConsoleCtrlHandler);
{$ENDIF}
  repeat
    FDispatcher.run(false);
    // After this call state is either ShutdownClient, or Running
    do_shutdown();
  until FState <> osRunning;
end;

{**
  Waits for answer to request with specified message id
}
function TORB.wait(id: ORBMessageID; timeout : long): Boolean;
var
  rec : IORBInvokeRecord;
begin
  result := False;
  rec := _get_invoke(id);
{$IFDEF HAVE_THREADS}
  if (rec <> nil) and rec.active then begin
    Assert(rec.callback() <> nil);
    result := rec.callback().waitfor(Self, id, _Result, timeout);
  end;
{$ELSE}
  if (rec = nil) or (rec.completed) then
    result := true
  else
    begin
      while not result do
        begin
          if (rec = nil) or (rec.completed) then
            begin
              result := true;
              break;
            end;
          FDispatcher.run(false);
          rec := get_invoke(id);
        end;
    end;
{$ENDIF}
end;

{**  Returns dispatcher}
function TORB.dispatcher(): ISelectDispatcher;
begin
  result := FDispatcher;
end;

{**  Removes invoke record from record list.}
procedure TORB.del_invoke(id: MessageID);
begin
{$IFDEF USE_ORB_CACHE}
  if FCacheRecord.id = id then begin
    FCacheUsed := false;
    FCacheRecord.clean;
  end;
{$ENDIF}
  FInvokesLock.wrlock();
  try
    FInvokes.erase(id);
  finally
    FInvokesLock.wrunlock();
  end; { try/finally }
end;

{**  Adds new invoke record}
procedure TORB.add_invoke(const inv: IORBInvokeRecord);
begin
{$IFDEF USE_ORB_CACHE}
  if inv <> FCacheRecord then
    FInvokes[inv.id] := inv;
{$ELSE}
  FInvokesLock.wrlock();
  try
    FInvokes[inv.id] := inv;
  finally
    FInvokesLock.wrunlock();
  end; { try/finally }
{$ENDIF}
end;

{**
  Return invoke record with specified message id
}
function TORB.get_invoke(mid: MessageID):IORBInvokeRecord;
begin
  result := nil;
{$IFDEF USE_ORB_CACHE}
  if FCacheRecord.id = mid then
    begin
      result := FCacheRecord;
      exit;
    end;
{$ENDIF}
  result := IORBInvokeRecord(FInvokes[mid]);
end;

function TOrb._get_invoke(id: ORBMessageID): IORBInvokeRecord;
begin
  result := id;
end;

{**
  Creates new invoke record. Returns cache record
  if it not used.
}
function TORB.create_invoke(msgid: MessageID): IORBInvokeRecord;
begin
{$IFDEF USE_ORB_CACHE}
  if FCacheUsed then
    result := TORBInvokeRecord.Create(msgid)
  else
    begin
      FCacheUsed := true;
      result := FCacheRecord;
    end;
{$ELSE}
  result := TORBInvokeRecord.Create(msgid);
{$ENDIF}
end;

{**
  Retrieves IOR and creates object with one
}
function TORB.string_to_object(const str: string): IORBObject;
var
  ior : TIOR;
begin
  if Pos('IOR:',Str)> 0 then
    begin
      ior := TIor.Create;
      ior.FORB := ORBInstance;
      if not ior.from_string(str) then
        ior.free
      else
        result := Ior_To_Object(ior);
    end
  else if Pos('http:',Str) > 0  then
    result := http_to_object(str)
  else if Pos('file:',str) > 0 then
    result := file_to_object(str)
  else
    result := nil;
end;

function TORB._get_interface(const obj: IORBObject): IInterfaceDef;
var
  req: IRequest;
  int : IUnknown;
begin
  result := nil;
  req := obj._request('_interface');
  req.get_result.argument.set_type(_tc_InterfaceDef);
  try
    req.invoke;
  except
  end;
  if req.get_environment.get_exception <> nil then
    dorb_throw(req.get_environment.get_exception);
  if not req.get_result.argument.to_object(int) then exit;
  result := TInterfaceDef._narrow(int as IORBObject);
end;

{**
  Creates object with specified IOR
}
function TORB.ior_To_Object(const Ior: IIOR): IORBObject;
var
  oa: IObjectAdaptor;
  temp : IORBObject;
begin
  result := nil;
  if ior.get_profile() = nil then exit;;
  temp := TORBObject.Create(ior);
  oa := get_oa(temp);
  if assigned(oa) and oa.is_local then
    result :=  oa.skeleton(temp)
  else
    result := temp;
end;

function TORB.invoke_async(const obj: IORBObject; const req: IORBRequest; const pr: IPrincipal;
  resp: Boolean; cb: IORBCallBack; orbmesid: ORBMessageID): ORBMessageID;
var
 oa : IObjectAdaptor;
 //rec : IORBInvokeRecord;
 ex : ISystemException;
 msgid: MessageID;
 id: ORBMessageID;
begin
  id := orbmesid;
  if id = nil then
    msgid := new_MessageID()
  else
    msgid := get_msgid(id);
  //set current msgid
  currentMsgId := msgid;
  if resp then
    begin
      if cb = nil then cb := TORBAsyncCallback.Create();
      if id = nil then id := create_invoke(msgid);
      id.init_invoke(self,obj,pr,req,true,cb);
      add_invoke(id);
    end;
  if not builtin_invoke(id,obj,req) then
    oa := get_oa(obj);
  if oa <> nil then
    begin
      if resp then
        id.oa(oa);
      oa.invoke(id,obj,req,pr,resp);
      result := id;
    end
  else
    begin
      ex := SystemException.Create();
      req.set_out_args(ex);
      answer_invoke(id,InvokeSysEx,obj,req);
      result := id;
    end;
end;

{**
  Returns new session unique id
**}
function TORB.new_MessageID(): MessageID;
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FTheIDLock);
  Inc(FMessageCounter);
  result := FMessageCounter;
end;

{**
  Adds Object Adaptor into list of adaptors
}
procedure TORB.register_oa(const oa: IObjectAdaptor; prio: integer);
begin
  FAdaptersLock.wrlock();
  try
    oa._AddRef;
    FAdapters.AddObject(IntToStr(prio),Pointer(oa));
    FAdapters.Sort;
  finally
    FAdaptersLock.wrunlock();
  end; { try/finally }
end;

{**
  Removes Object Adaptor from list of adaptors
}
procedure TORB.unregister_oa(const oa: IObjectAdaptor);
var
  i : integer;
begin
  FAdaptersLock.wrlock();
  try
    i := FAdapters.IndexOfObject(Pointer(oa));
    if i >= 0 then
      begin
        oa._Release;
        FAdapters.Delete(i);
      end;
  finally
    FAdaptersLock.wrunlock();
  end; { try/finally }
end;

{**
  Retrieves IOR using HTTP protocol and creates new object
}
function TORB.http_to_object(const str: string): IORBObject;
var
  i,total,allocated : longint;
  host,abspath : string;
  port : integer;
  addr : IInetAddress;
  trans : ITransport;
  buffer,bufptr : PChar;
  lstr : string;
const
  endblock : array[0..3] of byte = (13,10,13,10);
  blocksize = 256;
begin
  result := nil;
  if Pos('//',str) = 6 then
    begin
      lstr := copy(str,8,100);
      i := Pos('/',lstr);
      if i > 0 then
        begin
          abspath := PChar(Copy(lstr,i,100));
          host := PChar(Copy(lstr,1,i-1));
        end
      else
        begin
          abspath := '/';
          host := lstr;
        end;
      if Pos(':',host) > 0 then
        begin
          i := Pos(':',host);
          port := StrToInt(Copy(host,i+1,100));
          host := PChar(Copy(host,0,i-1));
        end
      else
        port := 80;
      addr := TInetAddress.Create(host,port,STREAM);
      if addr.resolve_ip() then
        begin
          trans := addr.make_transport;
          if trans.connect(addr) then
            begin
              if (trans.write(PChar('GET ')^,4) <> 4) or
                (trans.write(PChar(abspath)^,Length(abspath)) <> Integer(length(abspath))) or
                (trans.write(endblock,4) <> 4) then
                exit
              else
                begin
                  allocated := blocksize;
                  GetMem(Buffer,allocated);
                  try
                    bufptr := buffer;
                    total := 0;
                    while not trans.eof and not trans.bad do
                      begin
                        i := trans.read(bufptr^,blocksize);
                        if i > 0 then
                          begin
                            inc(total,i);
                            if i = blocksize then
                              begin
                                reallocmem(Buffer,allocated+blocksize);
                                inc(allocated,blocksize);
                                bufptr := Pointer(Integer(buffer)+total);
                              end
                            else
                              break;
                          end;
                      end;
                    trans := nil;  
                    if Total > 0 then
                      begin
                        if allocated < total then
                          begin
                            reallocmem(Buffer,total);
                            allocated := total;
                          end;
                        Buffer[total] := Chr(0);
                        result := string_to_object(PChar(Buffer));
                      end;
                  finally
                    freemem(buffer,allocated);
                  end;
                end;
            end;
        end;
    end;
end;

{**  Gets IOR from file and creates new object}
function TORB.file_to_object(const str: string): IORBObject;
var
  fl,l : integer;
  Buffer : PChar;
  lstr: string;
begin
  result := nil;
  if pos('//',str) > 0 then
    lstr := PChar(Integer(str)+Pos('//',str)+1)
  else
    lstr := str;
  fl := FileOpen(lstr,fmShareDenyNone);
  if fl >= 0 then
    try
      l := FileSeek(fl,0,2);
      FileSeek(fl,0,0);
      Buffer := PChar(AllocMem(l + 1));
      if FileRead(fl, Buffer^, l) = l then
        result := string_to_object(Buffer);
    finally
      FileClose(fl);
    end;
end;

{**  Returns first adaptor what service specified object}
function TORB.get_oa(const obj: IORBObject): IObjectAdaptor;
var
  i : integer;
  local : Boolean;
begin
  local := is_local(obj);
  FAdaptersLock.rdlock();
  try
    for i := Pred(FAdapters.Count) downto 0 do
      with IObjectAdaptor(Pointer(FAdapters.Objects[i])) do
      if (is_local = local) and  has_object(obj) then
        begin
          result := IObjectAdaptor(Pointer(FAdapters.Objects[i]));
          break;
        end;
  finally
    FAdaptersLock.rdunlock();
  end; { try/finally }
end;

{**  Gets reply on request with specified message id}
function TORB.get_invoke_reply (id: ORBMessageID;out obj: IORBObject; out req: IORBRequest): TInvokeStatus;
var
  irec : IORBInvokeRecord;
  res: Boolean;
begin
  result := orbtypes.InvokeOK;
  irec := _get_invoke(id);
  assert(irec<>nil);
  res := irec.get_answer_invoke(result,obj,req);
  assert(res);
  del_invoke(irec.id());
  currentMsgId := 0;
end;

function TORB.resolve_initial_reference(const id: ObjectID): IORBObject;
var
  i : integer;
  tmp : IORBObject;
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FInitRefsLock);
  result := nil;
  i := FInitRefs.IndexOf(id);
  if i < 0 then
    begin
      if id = 'ImplementationRepository' then
        begin
          tmp := TImplRepository_impl.Create();
          tmp._AddRef;
          FInitRefs.AddObject(id,Pointer(tmp));
        end
      else if id = 'LocalImplementationRepository' then
        begin
          tmp := resolve_initial_reference('ImplementationRepository');
          if is_impl(tmp) then
            tmp := TImplRepository_impl.Create();
          tmp._AddRef;
          FInitRefs.AddObject(id,Pointer(tmp));
        end
      else if (id = 'RootPOA') or (id = 'POACurrent') then begin // RootPOA initializes POACurrent
        if the_root_poa = nil then
          TPOA_impl.Create(self);
      end
      else if id = 'PrincipalCurrent' then begin
        tmp := TPrincipalCurrent_impl.Create;
        tmp._AddRef;
        FInitRefs.AddObject(id,Pointer(tmp));
      end;
      i := FInitRefs.IndexOf(id);
    end;
  if i >= 0 then
    result := IORBObject(Pointer(FInitRefs.Objects[i]));
end;

function TORB.register_value_factory(const id: RepositoryID; const factory: IValueFactory): IValueFactory;
begin
  result := lookup_value_factory(id);
  factory._AddRef;
  if result <> nil then
    begin
      result._Release;
      FValueFactories.Delete(FValueFactories.IndexOf(id));
    end;
  FValueFactories.AddObject(id,Pointer(factory));
end;

function TORB.lookup_value_factory(const id: RepositoryID): IValueFactory;
var
  i : integer;
begin
  result := nil;
  i := FValueFactories.IndexOf(id);
  if i>=0 then
    result := IValueFactory(Pointer(FValueFactories.Objects[i]))
end;

procedure TORB.unregister_value_factory(const id: RepositoryID);
var
  i : integer;
begin
  i := FValueFactories.IndexOf(id);
  if i < 0 then exit;
  IValueFactory(Pointer(FValueFactories.Objects[i]))._Release;
  FValueFactories.Delete(i);
end;

function TOrb.create_environment: IEnvironment;
begin
  result := TEnvironment.Create();
end;

function TOrb.object_to_string(const obj: IORBObject): string;
var
  ior : IIOR;
begin
  if obj = nil then
    begin
      ior := TIOR.Create();
      result := ior.stringify;
    end;
  if obj.ior = nil then
    dorb_throw(SystemException.Create_type(st_MARSHAL));
  result := obj.ior.stringify;
end;

procedure TOrb._destroy;
var
  i: integer;
  sync: ISynchronized;
begin
  if FState = osRunning then
    shutdown(true);
  // remove all profile
  FIORTempl.clean();
  FIORTempl := nil;
  // remove references
  sync := TSynchronized.Create(FInitRefsLock);
  for i := 0 to FInitRefs.Count - 1 do
    IORBObject(Pointer(FInitRefs.Objects[i]))._Release;
  FInitRefs.Clear;
  ORBInstance := nil;
end;

function TOrb.create_abstract_interface_tc(const id: RepositoryID;
  const name: Identifier): ITypeCode;
begin
  result := nil;
  assert(false,'not imlplemented');
end;

function TOrb.create_alias_tc(const id: RepositoryID;
  const name: Identifier; const original_type: ITypeCode): ITypeCode;
begin
  result := TTypeCode.create_alias_tc(id,name,original_type)
end;

function TOrb.create_array_tc(const length: _ulong;
  const element_type: ITypeCode): ITypeCode;
begin
  result := TTypeCode.create_array_tc(element_type,length)
end;

function TOrb.create_enum_tc(const id: RepositoryID;
  const name: Identifier; const members: EnumMemberSeq): ITypeCode;
begin
  result := TTypeCode.create_enum_tc(id,name,members)
end;

function TOrb.create_exception_tc(const id: RepositoryID;
  const name: Identifier; const members: StructMemberSeq): ITypeCode;
begin
  result := nil;
  assert(false,'not imlplemented');
end;

function TOrb.create_fixed_tc(const digits: _ushort;
  const scale: short): ITypeCode;
begin
  result := nil;
  assert(false,'not imlplemented');
end;

function TOrb.create_interface_tc(const id: RepositoryID;
  const name: Identifier): ITypeCode;
begin
  result := TTypeCode.create_interface_tc(id,name)
end;

procedure TOrb.create_list(const count: Integer; out new_list: INVList);
begin
  if count < 0 then
    dorb_throw(SystemException.Create_type(st_BAD_PARAM) as IORBException);
  new_list := TNVList.Create(count);
end;

function TOrb.create_native_tc(const id: RepositoryID;
  const name: Identifier): ITypeCode;
begin
  result := nil;
  assert(false,'not imlplemented');
end;

procedure TOrb.create_operation_list(const oper: IOperationDef;
  out new_list: INVList);
var
  params: ParDescriptionSeq;
  i : integer;
begin
  new_list := TNVList.Create();
  params := oper.get_params();
  for i := 0 to Pred(Length(params)) do
    case params[i].mode of
      PARAM_IN: new_list.add(ARG_IN).argument.set_type(params[i]._type);
      PARAM_OUT: new_list.add(ARG_OUT).argument.set_type(params[i]._type);
      PARAM_INOUT: new_list.add(ARG_INOUT).argument.set_type(params[i]._type);
    end;
end;

function TOrb.create_policy(_type: PolicyType; const val: IAny): IPolicy;
var
  v : longint;
begin
  result := nil;
  if not val.to_long(v) then
    raise TPolicyError.Create(BAD_POLICY_TYPE);
  case _type of
    THREAD_POLICY_ID : result := TThreadPolicy.Create(_type,ThreadPolicyValue(v));
    LIFESPAN_POLICY_ID : result := TLifespanPolicy.Create(_type,LifeSpanPolicyValue(v));
    ID_UNIQUENESS_POLICY_ID : result := TIdUniquenessPolicy.Create(_type,IdUniquenessPolicyValue(v));
    ID_ASSIGNMENT_POLICY_ID : result := TIdAssignmentPolicy.Create(_type,IdAssignmentPolicyValue(v));
    IMPLICIT_ACTIVATION_POLICY_ID : result := TImplicitActivationPolicy.Create(_type,ImplicitActivationPolicyValue(v));
    SERVANT_RETENTION_POLICY_ID : result := TServantRetentionPolicy.Create(_type,ServantRetentionPolicyValue(v));
    REQUEST_PROCESSING_POLICY_ID : result := TRequestProcessingPolicy.Create(_type,RequestProcessingPolicyValue(v));
    else
      raise TPolicyError.Create(BAD_POLICY_TYPE);
  end;
end;

function TOrb.create_recursive_tc(const id: RepositoryID): ITypeCode;
begin
  assert(false,'not imlplemented');
  result := nil;
end;

function TOrb.create_sequence_tc(const bound: _ulong;
  const element_type: ITypeCode): ITypeCode;
begin
  result := TTypeCode.create_sequence_tc(element_type,bound)
end;

function TOrb.create_string_tc(const bound: _ulong): ITypeCode;
begin
  result := nil;
  assert(false,'not imlplemented');
end;

function TOrb.create_struct_tc(const id: RepositoryID;
  const name: Identifier; const members: StructMemberSeq): ITypeCode;
begin
  result := TTypeCode.create_struct_tc(id,name,members)
end;

function TOrb.create_union_tc(const id: RepositoryID;
  const name: Identifier; const discriminator_type: ITypeCode;
  const members: UnionMemberSeq): ITypeCode;
begin
  result := TTypeCode.create_union_tc(id,name,discriminator_type,members)
end;

function TOrb.create_value_box_tc(const id: RepositoryID;
  const name: Identifier; const boxed_type: ITypeCode): ITypeCode;
begin
  assert(false,'not imlplemented');
  result := nil;
end;

function TOrb.create_value_tc(const id: RepositoryID;
  const name: Identifier; const type_modifier: TValueModifier;
  const concrete_base: ITypeCode;
  const members: ValueMemberSeq): ITypeCode;
begin
  result := TTypeCode.create_value_tc(id,name,type_modifier,
    concrete_base,members)
end;

function TOrb.create_wstring_tc(const bound: _ulong): ITypeCode;
begin
  assert(false,'not imlplemented');
  result := nil;
end;

procedure TOrb.get_default_context(out ctx: IContext);
begin
  
end;

procedure TOrb.get_next_response(out req: IRequest);
begin
  assert(false,'not imlplemented');
end;

function TOrb.get_service_information(const service_type: ServiceType;
  out service_information: ServiceInformation): Boolean;
begin
  result := false;
  assert(false,'not imlplemented');
end;

function TOrb.list_initial_references: ObjectIDList;
begin
  result := nil;
  assert(false,'not imlplemented');
end;

procedure TOrb.perform_work;
begin
  assert(false,'not imlplemented');
end;

function TOrb.poll_next_response: Boolean;
begin
  result := false;
  assert(false,'not imlplemented');
end;

procedure TOrb.send_multiple_requests_deferred(const req: RequestSeq);
begin
  assert(false,'not imlplemented');
end;

procedure TOrb.send_multiple_requests_oneway(const req: RequestSeq);
begin
  assert(false,'not imlplemented');
end;

function TOrb.work_pending: Boolean;
begin
  result := false;
  assert(false,'not imlplemented');
end;

function TOrb.create_named_value: TNamedValue;
begin
  result.name := '';
  result.argument := TAny.Create;
  result.len := 0;
  result.flags := 0;
end;

function TOrb.get_current_invoke_rec: IORBInvokeRecord;
begin
  if currentMsgId <> 0 then
    result := get_invoke(currentMsgId);
end;

function TOrb.get_current_msgid: MessageID;
begin
  result := currentMsgId;
end;

function TOrb.get_Properties: IProperties;
begin
  result := FProperties;
end;

procedure TOrb.set_Properties(const props: IProperties);
begin
  FProperties := props;
end;

procedure TOrb.validateState;
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FShutdownComplete);
  // The ORB destroys this object, so it's an initialization error
  // if the this operation is called after ORB destruction
  if FState = osDestroyed then
    dorb_throw(st_INITIALIZE, MinorORBDestroyed, COMPLETED_NO);
  if (FState = osServerShutdown) or (FState = osClientShutdown) then
    dorb_throw(st_BAD_INV_ORDER, MinorShutdownCalled, COMPLETED_NO);
  if FState = osNotRunning then begin
    // Remember the main thread id
    FMainThread := TORBThread.CurrentThread();
    // Set the state to StateRunning
    FState := osRunning;
  end;
end;

procedure TOrb.blockServerShutdownComplete;
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FShutdownComplete);
  // Wait for the server side to shutdown. Note that the client side
  // shutting down or the state being destroyed also implies that
  // the server side has shutdown
  while FState = osRunning do
    FShutdownComplete.wait();
end;

//********************************************************************
//  TORBInvokeRecord
//********************************************************************
constructor TORBInvokeRecord.Create(msgId: MessageID);
begin
  inherited Create();
  FMessageID := msgId;
  FActive := true;
end;

function TORBInvokeRecord.active: boolean;
begin
  result := FActive;
end;

procedure TORBInvokeRecord.deactivate;
begin
  FActive := false;
end;

function TORBInvokeRecord.get_InvokeHint: Pointer;
begin
  result := FInvHint;
end;

function TORBInvokeRecord.get_RequestHint: Pointer;
begin
  result := FReqHint;
end;

procedure TORBInvokeRecord.set_InvokeHint(hint: Pointer);
begin
  FInvHint := hint;
end;

procedure TORBInvokeRecord.set_RequestHint(hint: Pointer);
begin
  FReqHint := hint;
end;

procedure TORBInvokeRecord.init_invoke(const orb: IORB; const obj: IORBObject;
  const pr: IPrincipal; const req: IORBRequest; resp: Boolean; const cb: IORBCallBack);
begin
  FORB:= orb;
  FObject := obj;
  FRequest := req;
  FResponse_req := resp;
  FCallBack := cb;
  FAdapter := nil;
  FType := RequestInvoke;
  FHaveResult := False;
  FTag := nil;
  FAddress := nil;
  FActive := true;
  FPrincipal := pr;
end;

procedure TORBInvokeRecord.init_bind(const orb: IORB; const repoid: string;
  const tag: IObjectTag; const addr: IAddress; const cb: IORBCallBack;
  const oa: IObjectAdaptor);
begin
  FORB:= orb;
  FObject := nil;
  FRequest := nil;
  FResponse_req := TRUE;
  FCallBack := cb;
  FAdapter := oa;
  FTag := tag;
  FAddress := addr;
  FHaveResult := False;
  FType := RequestBind;
  FActive := true;
end;

function TORBInvokeRecord.callback(): IORBCallBack;
begin
  result := FCallBack;
end;

function TORBInvokeRecord.get_principal: IPrincipal;
begin
  result := FPrincipal;
end;

{**
  Returns record to uninitialized state
**}
procedure TORBInvokeRecord.clean();
begin
  FMessageID := 0;
  FObject := nil;
  FRequest := nil;
  FCallback := nil;
  FAdapter := nil;
  FHaveResult := false;
  FORB := nil;
  FStat := InvokeOK;
  FTag := nil;
  FAddress := nil;
  FActive := false;
end;

procedure TORBInvokeRecord.set_answer_bind(const state: TLocateStatus;const obj: IORBObject);
begin
  FLocateStatus := state;
  if state = LocateHere then
    FObject := obj;
  FHaveResult := true;
  Assert(FCallback <> nil);
  FCallback.notify(FORB, Self, _Bind);
end;

{**  Sets answer on invoke}
procedure TORBInvokeRecord.set_answer_invoke(const state: TInvokeStatus;const obj: IORBObject;const req: IORBRequest);
var
  ex: ISystemException;
begin
  FStat := state;
  if state = orbtypes.InvokeForward then
    FObject := obj
  else if state in [InvokeOk,InvokeUserEx,InvokeSysEx] then
    begin
      if not FRequest.copy_out_args(req) then
        begin
          ex := SystemException.Create_type(st_MARSHAL,0,COMPLETED_MAYBE);
          FRequest.set_out_args(ex);
          Fstat := InvokeSysEx;
        end;
    end;
  FHaveResult := true;
  Assert(FCallback <> nil);
  FCallback.notify(FORB, Self, _Invoke);
end;

{**  Returns request type of record}
function TORBInvokeRecord.request_type: TRequestType;
begin
  result := FType;
end;

{**  Sets Object Adaptor}
procedure TORBInvokeRecord.set_oa(const oa: IObjectAdaptor);
begin
  FAdapter := oa;
end;

{**
  Returns message id of record
**}
function TORBInvokeRecord.id: MessageID;
begin
  result := FMessageID;
end;

{**  TRUE if record contains answer on request}
function TORBInvokeRecord.completed(): Boolean;
begin
  result := FHaveResult;
end;

function TORBInvokeRecord.get_answer_bind(out state: TLocateStatus;out obj: IORBObject): Boolean;
begin
  result := false;
  if not self.FHaveResult then exit;
  obj := FObject;
  state := FLocateStatus;
  result := true;
end;

{**  Returns answer on request}
function TORBInvokeRecord.get_answer_invoke(out state: TInvokeStatus;out obj: IORBObject;out req: IORBRequest): Boolean;
begin
  result := false;
  if not FHaveResult then exit;
  obj := FObject;
  state := FStat;
  req := FRequest;
  result := true;
end;

{**  Returns Object Adaptor}
procedure TORBInvokeRecord.oa(oa: IObjectAdaptor);
begin
  FAdapter := oa;
end;

function TORBInvokeRecord.get_oa(): IObjectAdaptor;
begin
  result := FAdapter;
end;
{$ENDIF}

//***************************************************************
//  TObjectTag
//***************************************************************
constructor TObjectTag.Create(Name: string);
begin
  FName := Name;
end;

function TObjectTag.get_length: _ulong;
begin
  result := Length(FName);
end;

function TObjectTag.get_bytes: string;
begin
  result := FName;
end;

function TObjectTag.equals(const val: string): Boolean;
begin
  result := val = FName;
end;

{ TORBAsyncCallback }

constructor TORBAsyncCallback.Create;
begin
  inherited;
  FNotified := false;
  FCondMutex := TRecursiveMutex.Create;
  FCond := TCondVar.Create(FCondMutex);
end;

destructor TORBAsyncCallback.Destroy;
begin
  FCond.Free;
  FCondMutex.Free;
  inherited;
end;

procedure TORBAsyncCallback.notify(const orb: IORB; mid: ORBMessageID;
  event: TORBCallbackEvent);
begin
  FCondMutex.lock();
  try
    FNotified := true;
    FCond.broadcast();
  finally
    FCondMutex.unlock();
  end; { try/finally }
end;

function TORBAsyncCallback.waitfor(const orb: IORB; mid: ORBMessageID;
  event: TORBCallbackEvent; tmout: Integer): boolean;
begin
  FCondMutex.lock();
  try
    if FNotified then begin
      result := true;
      Exit;
    end;
    if tmout <> -1 then
      FCond.wait(tmout)
    else
      FCond.wait();
    result := FNotified;
  finally
    FCondMutex.unlock();
  end; { try/finally }
end;

{ TPrincipalCurrent }

class function TPrincipalCurrent._narrow(
  const obj: IORBObject): IPrincipalCurrent;
var
  p: Pointer;
begin
  if obj = nil then exit;
  p := obj.narrow_helper('IDL:omg.org/PrincipalCurrent:1.0');
  if p <> nil then
    result := IPrincipalCurrent(p);
end;

{ TPrincipalCurrent_impl }

function TPrincipalCurrent_impl.get_principal: IPrincipal;
var
  rec: IORBInvokeRecord;
begin
  rec := ORB_Instance.get_current_invoke_rec();
  if rec <> nil then
    result := rec.get_principal();
end;

function TPrincipalCurrent_impl.narrow_helper(const repoid: string): Pointer;
begin
  if repoid = 'IDL:omg.org/PrincipalCurrent:1.0' then
    begin
      result := Pointer(self as IPrincipalCurrent);
      exit;
    end;
  result := inherited narrow_helper(repoid);
end;

{ TProperties }

const
  cPropertyPrefix = 'mtdorb';

constructor TProperties.Create(props: TStrings);
var
  i: integer;
  str: string;

  procedure SetProperty(const AProp: string; var idx: integer);
  begin
    Inc(idx);
    if idx >= props.Count then Exit;
    set_property(cPropertyPrefix + AProp, props.Strings[idx]);
  end;

begin
  FProps := TStringList.Create;
  LoadDefaultProperties();
  if props = nil then Exit;
  i := 0;
  while i < props.Count do begin
    str := props.Strings[i];
    if str = '-ORBProperty' then begin
      if i + 1 < props.Count then
        FProps.Add(props.Strings[i + 1]);
      Inc(i);
    end
    else if str = '-ORBConnectionsLimit' then
      SetProperty('.orb.connections_limit', i)
    else if str = '-ORBThreadPoolTimeout' then
      SetProperty('.orb.thread_pool_timeout', i);
    Inc(i);
  end; { while }
end;

destructor TProperties.Destroy;
begin
  FProps.Free;
  inherited;
end;

function TProperties.get_Property(const key: string): string;
begin
  result := FProps.Values[key]
end;

procedure TProperties.set_Property(const key, value: string);
begin
  FProps.Values[key] := value;
end;

procedure TProperties.LoadDefaultProperties;
begin
  //Not implemented
end;

{$IFDEF LIBRARY}
exports
  ORB_Init,
  ORB_Instance;
{$ENDIF}
initialization
finalization
{$IFNDEF USELIB}
  ORBInstance := nil;
{$ENDIF}
end.
