// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit req;

interface

uses
  orb_int,orbtypes,req_int,code_int,Classes,except_int,env_int;

{$IFNDEF USELIB}

type

  {** Implements org.omg.CORBA.Request } 
  TRequest = class(TInterfacedObject,IRequest,IORBCallback)
  private
    FOperation : string;
    FObject : IORBObject;
    FMID: ORBMessageID;
    FContext : IContext;
    FResult : TNamedValue;
    FEnvironment : IEnvironment;
    FArguments : INVList;
    FCallback : IRequestCallback;
    FRequest: IORBRequest;
    FIceptReq: IORBObject;
    FFlags : _ulong;
    FInvokePending: Boolean;
  protected
    function waitfor(const orb: IORB; id: ORBMessageID; event: TORBCallbackEvent; tmout: Longint = -1): boolean;
    procedure notify(const orb: IORB; id: ORBMessageID; event: TORBCallbackEvent);
    //procedure callback(const orb: IORB; mid: MessageID; event: TORBCallbackEvent);
    function add_in_arg(const name: string): IAny;
    function add_inout_arg(const name: string): IAny;
    function add_out_arg(const name: string): IAny;
    function get_arguments(): INVList;
    function get_environment(): IEnvironment;
    function get_exceptions(): IInterfaceList;
    function get_context(): IContext;
    function get_contexts(): IInterfaceList;
    procedure invoke();
    function operation(): string;
    function return_value: IAny;
    procedure set_context(const ctx: IContext);
    procedure set_result(const any: IStaticAny);
    procedure set_return_type(const tc: ITypeCode);
    function get_result(): TNamedValue;
    function get_target(): IORBObject;
    procedure decode_user_exception();

    procedure get_response(block: Boolean = true);
    function poll_response(): Boolean;
    procedure send_oneway();
    procedure send_deferred(const cb: IRequestCallBack = nil);
  public
    constructor Create(const obj : IORBObject; const c : IContext; op: string;
      arguments: INVList; res: TNamedValue; flags: _ulong); overload;
    constructor Create(const obj : IORBObject; op: string); overload;
  end;

  TORBRequest = class(TInterfacedObject)
  protected
    FOperation : string;
    FServCnts: ServiceContextList;
  protected
    function context(): PServiceContextList;
    procedure set_context(const ctx: PServiceContextList);
    function op_name(): string; virtual;
    procedure set_out_args_ex(const ex: IORBException); virtual;
    procedure copy_servcntx(r: IORBRequest);
  end;

  TLocalRequest = class(TORBRequest,IORBRequest)
  private
    FResult : TNamedValue;
    FRequest: IRequest;
    FHaveResult: Boolean;
    FHaveException : Boolean;
  protected
    procedure add_in_arg(const any: IStaticAny);
    procedure add_inout_arg(const any: IStaticAny);
    procedure add_out_arg(const any: IStaticAny);
    function copy_out_args(const req: IORBRequest): Boolean;
    function get_in_args(const enc: IEncoder): Boolean; overload;
    function get_in_args(const list : INVList; out ctx: IContext): Boolean; overload;
    function get_in_args(const list : IStaticAnyList; out ctx: IContext): Boolean; overload;
    function get_out_args(const enc: IEncoder; out is_except: Boolean): Boolean; overload;
    function get_out_args(const res: IAny; const list: INVList; out ex: IORBException): Boolean; overload;
    function get_out_args(const res: IStaticAny; const list: IStaticAnyList; out ex: IORBException): Boolean; overload;
    procedure invoke();
    function op_name(): string; override;
    function req_type(): string;
    procedure set_result(const any: IStaticAny);
    function set_out_args(const dec: IDecoder; const exception : Boolean): Boolean; overload;
    function set_out_args(const res: IStaticAny; const list: IStaticAnyList): Boolean;  overload;
    function set_out_args(const ex: IORBException): Boolean; overload;
    function set_out_args(const res: IAny; const list: INVList): Boolean; overload;
  public
    constructor Create_req(const req: IRequest);
  end;

  TServerRequest = class(TInterfacedObject,IServerRequestBase,IServerRequest)
  private
    FArguments: INVList;
    FReq: IORBRequest;
    FObject: IORBObject;
    FMid: ORBMessageID;
    FAdaptor: IObjectAdaptor;
    FResult : IAny;
    FEnvironment : IEnvironment;
    FContext: IContext;
    FAborted : Boolean;
    FIceptReq: IORBObject;
    FCanceled : Boolean;
    function environment: IEnvironment;
  protected
    procedure set_exception(const exc: IORBException);
    function operation: string;
    function arguments(const lst: INVList): Boolean;
    procedure set_result(const any: IAny);
    function ctx: IContext;
    procedure set_out_args();
  public
    constructor Create(const req: IORBRequest; const obj: IORBObject; mid: ORBMessageID;
      const oa: IObjectAdaptor; pr: IPrincipal);
    destructor Destroy(); override;
  end;

  function CreateRequest(const obj : IORBObject; op: string): IRequest;
  function CreateRequest_arg(const obj : IORBObject; const c : IContext; op: string;
      arguments: INVList; res: TNamedValue; flags: _ulong): IRequest;

implementation

uses
  env,exceptions,intercept_int,intercept,throw;

//**************************************************************
// TRequest
//**************************************************************
constructor TRequest.Create(const obj : IORBObject; op: string);
var
  orb: IORB;
begin
  if obj = nil then dorb_throw(SystemException.Create_type(st_BAD_PARAM) as IORBException);
  if obj.ior = nil then dorb_throw(SystemException.Create_type(st_NO_IMPLEMENT) as IORBException);
  FObject := obj;
  FOperation := op;
  FContext := nil;
  orb := FObject.orb;
  FResult := orb.create_named_value;
  FEnvironment := orb.create_environment;
  FRequest := TLocalRequest.Create_req(self);
  //ClientInterceptor
  FIceptReq := TClientInterceptor._create_request(FObject, FRequest.op_name(), FRequest.context(), Self);
  FCallback := nil;
  orb.create_list(0,FArguments);
  FFlags := 0;
end;

constructor TRequest.Create(const obj : IORBObject; const c : IContext; op: string;
      arguments: INVList; res: TNamedValue; flags: _ulong);
var
  orb: IORB;
begin
  if obj = nil then dorb_throw(SystemException.Create_type(st_BAD_PARAM) as IORBException);
  if obj.ior = nil then dorb_throw(SystemException.Create_type(st_NO_IMPLEMENT) as IORBException);
  FObject := obj;
  FContext := c;
  FArguments := arguments;
  FResult := res;
  FFlags := flags;
  FOperation := op;
  orb := obj.orb;
  if FArguments = nil then
    orb.create_list(0,FArguments);
  if FResult.argument = nil then
    FResult := orb.create_named_value;
  FEnvironment := orb.create_environment;
  FRequest := TLocalRequest.Create_req(self);
  //ClientInterceptor
  FIceptReq := TClientInterceptor._create_request(FObject, FRequest.op_name(), FRequest.context(), Self);
  FCallback := nil;
end;

function TRequest.get_environment(): IEnvironment;
begin
  result := FEnvironment;
end;

procedure TRequest.notify(const orb: IORB; id: ORBMessageID;
  event: TORBCallbackEvent);
begin
  assert(event = _Invoke);
  if FCallback <> nil then
    begin
      FCallback.callback(self);
      FCallback := nil;
    end;
end;

function TRequest.operation: string;
begin
  result := FOperation
end;

procedure TRequest.decode_user_exception();
var
  ex: IORBException;
begin
  ex := UnknownUserException.EXCEPTION_DOWNCAST(FEnvironment.get_exception);
  if (ex <> nil) then
    begin
      assert(false,'not implemented');
    end;
end;

function TRequest.get_arguments(): INVList;
begin
  result := FArguments;
end;

function TRequest.add_in_arg(const name: string): IAny;
begin
  result := FArguments.add_item(name,ARG_IN).argument
end;

function TRequest.add_inout_arg(const name: string): IAny;
begin
  result := FArguments.add_item(name,ARG_INOUT).argument
end;

function TRequest.add_out_arg(const name: string): IAny;
begin
  result := FArguments.add_item(name,ARG_OUT).argument
end;

function TRequest.get_context: IContext;
begin
  result := FContext
end;

function TRequest.get_contexts: IInterfaceList;
begin
  assert(false,'not implemented');
end;

function TRequest.get_exceptions: IInterfaceList;
begin
  assert(false,'not implemented');
end;

function TRequest.get_result: TNamedValue;
begin
  result := FResult;
end;

function TRequest.get_target: IORBObject;
begin
  result := FObject
end;

procedure TRequest.invoke;
begin
  send_deferred;
  get_response;
end;

function TRequest.return_value: IAny;
begin
  Result := FResult.argument
end;

procedure TRequest.set_context(const ctx: IContext);
begin
  FContext := ctx
end;

procedure TRequest.set_result(const any: IStaticAny);
begin
  assert(false,'not implemented');
end;

procedure TRequest.set_return_type(const tc: ITypeCode);
begin
  FResult.argument.set_type(tc)
end;

procedure TRequest.get_response(block: Boolean);
var
  orb : IORB;
  rs: TInvokeStatus;
  obj : IORBObject;
  req: IORBRequest;
  s : ISystemException;
begin
  if not FInvokePending then exit;
  orb := FObject.orb;
  rs := InvokeOK;
  while FInvokePending do
    begin
      if block then
        orb.wait(FMid,0)
      else if not orb.wait(FMid,0) then
        exit;
      if not TClientInterceptor._exec_before_unmarshal(FIceptReq as ILWRequest,FEnvironment) then
        begin
          orb.cancel(Fmid);
          FInvokePending := false;
          if FCallback <> nil then FCallback.callback(self);
          FCallback := nil;
          exit;
        end;
      rs := orb.get_invoke_reply(FMid,obj,req);
      case rs of
        InvokeOK: begin
            FInvokePending := false;
            break;
          end;
        InvokeForward: begin
            FObject._forward(obj);
            FMid := orb.invoke_async(FObject,FRequest,nil,true);
            break;
          end;
        InvokeSysEx: begin
            if FObject.is_forwarded then
              begin
                s := SystemException.EXCEPTION_DOWNCAST(FEnvironment.get_exception);
                if (s.completed = COMPLETED_NO) and (s.extype in [st_COMM_FAILURE,st_TRANSIENT,st_OBJECT_NOT_EXIST]) then
                  begin
                    FObject._unforward;
                    FEnvironment.clear;
                    FMid := orb.invoke_async(FObject,FRequest,nil,true);
                    break;
                  end
              end;
            FInvokePending := false;
            break;
          end;
        InvokeUserEx: begin
            decode_user_exception();
            FInvokePending := false;
            break;
          end;
      end;
    end;
  if not TClientInterceptor._exec_after_marshal(FIceptReq as ILWRequest,FEnvironment) then
    begin
      FCallback := nil;
      exit;
    end;
  FCallback := nil;
  if not FInvokePending and (rs = InvokeSysEx) then
    FEnvironment.get_exception.throw;
end;

function TRequest.poll_response: Boolean;
begin
  get_response(false);
  result := FInvokePending;
end;

procedure TRequest.send_deferred(const cb: IRequestCallBack);
var
  orb: IORB;
begin
  FCallback := cb;
  if not TClientInterceptor._exec_initialize_request(FIceptReq as ILWRequest,FEnvironment) then
    begin
      if FCallback <> nil then
        FCallback.callback(self);
      FCallback := nil;
      exit;
    end;
  FInvokePending := true;
  orb := FObject.orb;
  FMid := orb.new_orbid;
  if assigned(FCallback) then
    orb.invoke_async(FObject,FRequest,nil,true,self,Fmid)
  else
    orb.invoke_async(FObject,FRequest,nil,true,nil,Fmid);
  if FInvokePending and not TClientInterceptor._exec_after_marshal(FIceptReq as ILWRequest,FEnvironment) then
    begin
      orb.cancel(Fmid);
      FInvokePending := false;
      if FCallback <> nil then
        FCallback.callback(self);
      FCallback := nil;
    end;
end;

procedure TRequest.send_oneway;
begin
  if not TClientInterceptor._exec_initialize_request(FIceptReq as ILWRequest,FEnvironment) then
      exit;
  FObject.orb.invoke_async(FObject,FRequest,nil,false);
  TClientInterceptor._exec_after_marshal(FIceptReq as ILWRequest,FEnvironment)
end;

//********************************************************************
// TORBRequest
//********************************************************************
function TORBRequest.context: PServiceContextList;
begin
  result := @FServCnts;
end;

procedure TORBRequest.copy_servcntx(r: IORBRequest);
begin
  FServCnts := r.context()^;
end;

function TORBRequest.op_name(): string;
begin
  result := FOperation;
end;

procedure TORBRequest.set_context(const ctx: PServiceContextList);
begin
  FServCnts := ctx^;
end;

procedure TORBRequest.set_out_args_ex(const ex: IORBException);
begin
  assert(false,'not implemented');
end;

//********************************************************************
// TLocalRequest
//********************************************************************
procedure TLocalRequest.add_in_arg(const any: IStaticAny);
begin
  assert(false,'not imlplemented');
end;

procedure TLocalRequest.add_out_arg(const any: IStaticAny);
begin
  assert(false,'not imlplemented');
end;

procedure TLocalRequest.add_inout_arg(const any: IStaticAny);
begin
  assert(false,'not imlplemented');
end;

function TLocalRequest.copy_out_args(const req: IORBRequest): Boolean;
var
  ex : IORBException;
begin
  result := false;
  if req = IORBRequest(Self) then
    begin
      result := true;
      exit;
    end;
  copy_servcntx(req);
  if not req.get_out_args(FRequest.get_result.argument,FRequest.get_arguments,ex) then exit;
  if assigned(ex) then
    begin
      FHaveResult := false;
      FHaveException := true;
      FRequest.get_environment.set_exception(ex);
    end
  else
    begin
      FHaveResult := true;
      FHaveException := false;
    end;
  result := true;
end;

constructor TLocalRequest.Create_req(const req: IRequest);
begin
  FRequest := req;
end;

function TLocalRequest.get_in_args(const enc: IEncoder): Boolean;
var
  args : INVList;
  i : integer;
begin
  result := false;
  args := FRequest.get_arguments;
  for i := 0 to Pred(args.count) do
    if (args.item(i).flags in [ARG_IN,ARG_INOUT]) and
      not args.item(i).argument.marshal(enc) then
        exit;
  if FRequest.get_context <> nil then
    FRequest.get_context.encode(enc,FRequest.get_contexts);
  result := true;
end;

function TLocalRequest.get_in_args(const list: INVList;
  out ctx: IContext): Boolean;
begin
  result := false;
  if not list.copy(FRequest.get_arguments,ARG_IN) then exit;
  ctx := FRequest.get_context;
  result := true;
end;

function TLocalRequest.get_in_args(const list: IStaticAnyList;
  out ctx: IContext): Boolean;
var
  args: INVList;
  nv : TNamedValue;
  i : integer;
begin
  result := false;
  args := FRequest.get_arguments;
  if list.Count <> args.count then exit;
  for i := 0 to Pred(list.count) do
    begin
      nv := args.item(i);
      if list[i].get_flags <> nv.flags then exit;
      if (list[i].get_flags in [ARG_IN,ARG_INOUT]) and
        not nv.argument.to_static_any(list[i]) then exit;
    end;
  ctx := FRequest.get_context;
  result := true;
end;


function TLocalRequest.get_out_args(const res: IStaticAny; const list: IStaticAnyList; out ex: IORBException): Boolean;
var
  e : IORBException;
  nvlist: INVList;
begin
  result := false;
  if FHaveException then
    begin
      e := FRequest.get_environment().get_exception;
      assert(e<>nil);
      ex := e.clone;
      result := true;
      exit;
    end;
  nvlist := FRequest.get_arguments;
  if (list.Count <> nvlist.Count) then
    exit;
  if FHaveResult and (res <> nil) then
    FRequest.get_result.argument.to_static_any(res);

end;

function TLocalRequest.get_out_args(const res: IAny; const list: INVList; out ex: IORBException): Boolean;
var
  e : IORBException;
  lst: INVList;
begin
  result := false;
  if FHaveException then
    begin
      e := FRequest.get_environment.get_exception();
      assert(e<>nil);
      ex := e.clone;
      result := true;
      exit;
    end;
  ex := nil;
  lst := FRequest.get_arguments;
  if lst.Count <> list.Count then
    exit;
  if FHaveResult then
    FResult := FRequest.get_result;

  result := true;
end;

function TLocalRequest.get_out_args(const enc: IEncoder; out is_except: Boolean): Boolean;
var
  args : INVList;
  i : integer;
begin
  result := false;
  if FHaveException then
    begin
      is_except := true;
      FRequest.get_environment.get_exception.encode(enc);
    end
  else
    begin
      is_except := false;
      if FHaveResult and not FRequest.get_result.argument.marshal(enc) then exit;
      args := FRequest.get_arguments;
      for i := 0 to Pred(args.count) do
        if (args.item(i).flags in [ARG_OUT,ARG_INOUT]) and
          not args.item(i).argument.marshal(enc) then
            exit;
    end;
  result := true;
end;

procedure TLocalRequest.invoke;
begin
  assert(false,'not imlplemented');
end;

function TLocalRequest.op_name: string;
begin
  result := FRequest.operation;
end;

function TLocalRequest.req_type: string;
begin
  result := 'local';
end;

function TLocalRequest.set_out_args(const ex: IORBException): Boolean;
begin
  result := true;
  FHaveResult := false;
  FHaveException := True;
  FRequest.get_environment.set_exception(ex.clone);
end;

function TLocalRequest.set_out_args(const dec: IDecoder;
  const exception: Boolean): Boolean;
var
  e : IORBException;
  tc: ITypeCode;
  args : INVList;
  i : integer;
begin
  result := false;
  if exception then
    begin
      FHaveResult := false;
      FHaveException := true;
      e := ORBException.decode(dec);
      FRequest.get_environment.set_exception(e);
    end
  else
    begin
      FHaveResult := true;
      FHaveException := false;
      tc := FRequest.get_result.argument.tc;
      if not FRequest.get_result.argument.demarshal(tc,dec) then exit;
      args := FRequest.get_arguments;
      for i := 0 to Pred(args.count) do
        if args.item(i).flags in [ARG_OUT,ARG_INOUT] then
          begin
            tc := args.item(i).argument.tc;
            if not args.item(i).argument.demarshal(tc,dec) then exit;
          end;
    end;
  result := true;
end;

function TLocalRequest.set_out_args(const res: IAny;
  const list: INVList): Boolean;
begin
  FHaveException := false;
  if res <> nil then
    begin
      FHaveResult :=  true;
      FRequest.get_result.argument.put_any(res);
    end;
  result := FRequest.get_arguments.copy(list,ARG_OUT or ARG_INOUT);
end;

function TLocalRequest.set_out_args(const res: IStaticAny;
  const list: IStaticAnyList): Boolean;
var
  args: INVList;
  nv: TNamedValue;
  i : integer;
begin
  result := false;
  args := FRequest.get_arguments;
  if args.count <> list.count then exit;
  if assigned(res) then
    FRequest.get_result.argument.from_static_any(res);
  for i := 0 to Pred(args.count) do
    begin
      nv := args.item(i);
      if nv.flags <> list[i].get_flags then exit;
      if (nv.flags in [ARG_OUT,ARG_INOUT] ) and
        not nv.argument.from_static_any(list[i]) then exit;
    end;
  result := true;
end;

procedure TLocalRequest.set_result(const any: IStaticAny);
begin
  assert(false,'not imlplemented');
end;

//*********************************************************************
//  TServerRequest
//*********************************************************************
constructor TServerRequest.Create(const req: IORBRequest;
  const obj: IORBObject; mid: ORBMessageID; const oa: IObjectAdaptor; pr: IPrincipal);
begin
  FEnvironment := TEnvironment.Create(pr);
  FReq := req;
  FObject := obj;
  FAborted := False;
  FMid := mid;
  FAdaptor := oa;
  FArguments := nil;
  FCanceled := false;
  //ServerIntercept
  FIceptReq := TServerInterceptor._create_request(FObject, req.op_name(), req.context(), Self);
end;

destructor TServerRequest.Destroy();
var
  stat : TInvokeStatus;
begin
  if not FCanceled then
    begin
      //ServerIntercept
      if (FIceptReq <> nil) and not FAborted and not (TServerInterceptor._exec_before_marshal(FIceptReq as ILWServerRequest, FEnvironment)) then begin
        FAborted := true;
      end;
      // copy back out args and result into ORBRequest
      set_out_args();
      stat := InvokeOK;
      if FEnvironment.get_exception <> nil then
        begin
          if SystemException.EXCEPTION_DOWNCAST(FEnvironment.get_exception) <> nil then
            stat := InvokeSysEx
          else
            stat := InvokeUserEx;
        end;
      FAdaptor.answer_invoke(FMid,FObject,FReq,stat);
      //ServerIntercept
      if (FIceptReq <> nil) and not FAborted and not (TServerInterceptor._exec_finish_request(FIceptReq as ILWServerRequest, FEnvironment)) then begin
        // XXX how to abort request here ??
      end;
    end;
  inherited;
end;

procedure TServerRequest.set_out_args();
begin
  if FEnvironment.get_exception <> nil then
    FReq.set_out_args(FEnvironment.get_exception)
  else if FArguments <> nil then
    begin
//      if not FReq.set_out_args(FResult,)

    end
end;

function TServerRequest.environment: IEnvironment;
begin
  result := FEnvironment;
end;

procedure TServerRequest.set_exception(const exc: IORBException);
begin
  FEnvironment.set_exception(exc);
  FResult := nil;
end;

function TServerRequest.operation: string;
begin
  result := freq.op_name;
end;

procedure TServerRequest.set_result(const any: IAny);
begin
  FResult := any;
end;

function TServerRequest.arguments(const lst: INVList): Boolean;
begin
  result := false;
  assert(FArguments = nil);
  FArguments := lst;
  if not TServerInterceptor._exec_initialize_request(FIceptReq as ILWServerRequest, FEnvironment) then
    begin
      FAborted := true;
      exit;
    end;
  if not FReq.get_in_args(FArguments,FContext) then
    begin
      set_exception(SystemException.Create_type(st_MARSHAL) as IORBException);
      FAborted := true;
      exit;
    end;
  if not TServerInterceptor._exec_after_unmarshal(FIceptReq as ILWServerRequest, FEnvironment) then
    begin
      FAborted := true;
      exit;
    end;
  result := true;
end;

function TServerRequest.ctx: IContext;
begin
  result := FContext;
end;

function TRequest.waitfor(const orb: IORB; id: ORBMessageID;
  event: TORBCallbackEvent; tmout: Integer): boolean;
begin
  assert(false,'not imlplemented');
  result := false;
end;

//*********************************************************************
function CreateRequest(const obj : IORBObject; op: string): IRequest;
begin
  result := TRequest.Create(obj,op)
end;

function CreateRequest_arg(const obj : IORBObject; const c : IContext; op: string;
      arguments: INVList; res: TNamedValue; flags: _ulong): IRequest;
begin
  result := TRequest.Create(obj,c,op,arguments,res,flags);
end;

{$IFDEF LIBRARY}
exports
  CreateRequest,
  CreateRequest_arg;
{$ENDIF}
{$ELSE}
  function CreateRequest(const obj : IORBObject; op: string): IRequest; external MTDORB_Library_Name;
  function CreateRequest_arg(const obj : IORBObject; const c : IContext; op: string;
      arguments: INVList; res: TNamedValue; flags: _ulong): IRequest; external MTDORB_Library_Name;
implementation
{$ENDIF}
end.
